use crate::prelude::Host;
use std::{fmt, str::FromStr};

mod error;
pub use error::Error;

#[derive(Clone, Debug, PartialEq)]
/// The full request as sent by the `Sender` and received by the `Receiver`
pub struct Request {
    /// The username of the sender
    pub user: String,
    /// The fully qualified domain name of the sending server
    pub host: Host,
    /// The message body
    pub message: String,
}

impl fmt::Display for Request {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "misfin://{}@{} {}\r\n",
            self.user, self.host, self.message
        )
    }
}

impl FromStr for Request {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Some((user, message)) = s.split_once(' ') {
            let Some(message) = message.strip_suffix("\r\n").map(ToString::to_string) else {
                return Err(Error::Malformed);
            };
            if message.is_empty() {
                return Err(Error::EmptyMessage);
            }
            if let Some((user, host)) = user.rsplit_once('@') {
                if host.is_empty() {
                    return Err(Error::EmptyHost);
                } else if user == "misfin://" {
                    return Err(Error::EmptyUser);
                }
                let host = host.parse()?;
                let Some(user) = user.strip_prefix("misfin://").map(ToString::to_string) else {
                    return Err(Error::Malformed);
                };
                Ok(Request {
                    user,
                    host,
                    message,
                })
            } else {
                Err(Error::MissingSeparator)
            }
        } else {
            Err(Error::MissingSeparator)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::prelude::ParseHostError;

    const REQ_STR: &'static str = "misfin://john@misfin.example.com Anyone seen Jane?\r\n";

    fn req() -> Request {
        Request {
            user: "john".to_string(),
            host: Host {
                subdomain: Some("misfin".into()),
                domain: "example".into(),
                tld: "com".into(),
            },
            message: "Anyone seen Jane?".to_string(),
        }
    }

    #[test]
    fn to_string() {
        assert_eq!(req().to_string().as_str(), REQ_STR,);
    }

    #[test]
    fn parse() {
        assert_eq!(REQ_STR.parse::<Request>().unwrap(), req());
    }

    #[test]
    fn parse_missing_sep() {
        let req = "misfin://john@example.comHelloWorld!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::MissingSeparator));
    }

    #[test]
    fn parse_malformed() {
        let req = "misfin://john@example.com Hello World!\n".parse::<Request>();
        assert_eq!(req, Err(Error::Malformed));
        let req = "mail://john@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::Malformed));
    }

    #[test]
    fn parse_empty_user() {
        let req = "misfin://@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyUser));
    }

    #[test]
    fn parse_empty_host() {
        let req = "misfin://john@ Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyHost));
    }

    #[test]
    fn parse_empty_msg() {
        let req = "misfin://john@example.com \r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyMessage));
    }

    #[test]
    fn parse_host_illegal_whitespace() {
        let req = "misfin://john@example\tfairy.com Hello World!\r\n".parse::<Request>();
        assert_eq!(
            req,
            Err(Error::ParseHostError(ParseHostError::IllegalWhitespace))
        );
    }
}
