use crate::fingerprint::Fingerprint;
use rustls::{
    client::{ServerCertVerified, ServerCertVerifier},
    Certificate,
};
use std::sync::{Arc, Mutex};

/// An item which stores known certificates
pub trait CertificateStore: Send + Sync {
    fn get(&self, host: &str) -> Option<String>;
    fn insert(&mut self, host: &str, fingerprint: &str);
}

#[derive(Debug)]
/// A verifier is used to verify certificates sent by the receiving server
/// during the tls handshake.
pub struct Verifier<S: CertificateStore> {
    /// An item which serves as storage for certificates
    pub store: Arc<Mutex<S>>,
}

impl<S: CertificateStore> ServerCertVerifier for Verifier<S> {
    fn verify_server_cert(
        &self,
        end_entity: &Certificate,
        _intermediates: &[Certificate],
        server_name: &rustls::ServerName,
        _scts: &mut dyn Iterator<Item = &[u8]>,
        _ocsp_response: &[u8],
        _now: std::time::SystemTime,
    ) -> Result<ServerCertVerified, rustls::Error> {
        let fp = end_entity
            .fingerprint()
            .map_err(|e| rustls::Error::General(e.to_string()))?;
        let name = match server_name {
            rustls::ServerName::DnsName(n) => n.as_ref().to_string(),
            rustls::ServerName::IpAddress(ip) => ip.to_string(),
            _ => todo!(),
        };
        if let Some(fingerprint) = match server_name {
            rustls::ServerName::DnsName(n) => self.store.lock().unwrap().get(n.as_ref()),
            rustls::ServerName::IpAddress(ip) => self.store.lock().unwrap().get(&ip.to_string()),
            _ => todo!(),
        } {
            if fingerprint == fp.1 && name == fp.0 {
                return Ok(ServerCertVerified::assertion());
            }
        } else {
            // todo: need a way to update `self.store`. Probably will require
            // an Arc<Mutex<T>> for interior mutability.
            // UPDATE: Now wrapped in Arc<Mutex<T>>
        }
        Err(rustls::Error::General(
            "Unrecognized certificate".to_string(),
        ))
    }
}

impl<T: CertificateStore> Verifier<T> {
    pub fn new(store: T) -> Self {
        Self {
            store: Arc::new(Mutex::new(store)),
        }
    }
}
